//Code for master device (ESP32) with file storage (SPIFFS) + BLE functions.
//The code is not final and it's still under development

#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <Wire.h>
#include <M5Stack.h>
#include "FS.h"
#include "SPIFFS.h"


BLEServer *pServer = NULL;
BLECharacteristic * pTxCharacteristic;
bool deviceConnected = false;
bool oldDeviceConnected = false;
String txValue;
String sensorName = "Temperature";
byte sensorAddress = 8;

char temperature_arr[6];    //array to convert temperature into string
float temperature;
int data_size;
int t = 2;
bool data_present = false;
bool send_saved = false;
bool send_RT = false;

//#define FORMAT_SPIFFS_IF_FAILED true //set this to true only when you run the SPIFFS for the time otherwise set it to false
#define FORMAT_SPIFFS_IF_FAILED false

#define pin 33 //battery status pin
#define pin_green 23 // green LED
#define pin_red 19 // red LED
#define gnd 25 // virtual ground
//#define vcc 21 // virtual vcc

float pinValue;
float pinVoltage;

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID           "6e400001-b5a3-f393-e0a9-e50e24dcca9e" // UART service UUID
#define CHARACTERISTIC_UUID_RX "6e400003-b5a3-f393-e0a9-e50e24dcca9e"
#define CHARACTERISTIC_UUID_TX "6e400002-b5a3-f393-e0a9-e50e24dcca9e"


class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string rxValue = pCharacteristic->getValue();

      if (rxValue.length() > 0) {
        Serial.println("*********");
        Serial.print("Received Value: ");
        for (int i = 0; i < rxValue.length(); i++)
          Serial.print(rxValue[i]);

        Serial.println();
        Serial.println("*********");
      }
    }
};


File file = SPIFFS.open("/sensorData.txt");


void setup() {

  M5.begin();  //Init M5Core. M5Core
  M5.Power.begin(); //Init Power module.
  M5.Lcd.setTextColor(YELLOW);  //Set the font color to yellow.
  M5.Lcd.setTextSize(2);  //Set the font size 2
  M5.Lcd.setCursor(65, 10); //Move the cursor position to (65, 10)
  M5.Lcd.println("LAMAP & FabLabSU"); //The screen prints the formatted string and wraps the line.
  delay(3000);
  
  Wire.begin();
  Serial.begin(115200);
  pinMode(pin_green, OUTPUT);
  pinMode(pin_red, OUTPUT);
  pinMode(gnd, OUTPUT);
  digitalWrite(gnd, LOW); //force pin 25 to be ground


  // Create the BLE Device
  BLEDevice::init("Arduino ESP32 M5");

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pTxCharacteristic = pService->createCharacteristic(
                    CHARACTERISTIC_UUID_TX,
                    BLECharacteristic::PROPERTY_NOTIFY
                  );
                      
  pTxCharacteristic->addDescriptor(new BLE2902());

  BLECharacteristic * pRxCharacteristic = pService->createCharacteristic(
                       CHARACTERISTIC_UUID_RX,
                      BLECharacteristic::PROPERTY_WRITE
                    );

  pRxCharacteristic->setCallbacks(new MyCallbacks());

  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  
  Serial.println("Waiting a client connection to notify...");
  
  M5.Lcd.clear(BLACK);
  M5.Lcd.setCursor(0, 35);
  M5.Lcd.println("Waiting a client connection to notify...");


  if(!SPIFFS.begin(FORMAT_SPIFFS_IF_FAILED))
    {
        Serial.println("SPIFFS Mount Failed");
        return;
    }
}

void loop() {

  M5.update(); //Read the press state of the key. A, B, C

  //------------------------------------------------------------------------------------------------------------------------------------------------------
  /*
  //battery monitor using LEDs
  pinValue = analogRead(pin); //raw digital values
  pinVoltage  = (pinValue/4095)*3.3; //voltage in V

  if (pinVoltage >= 3.1)
  {
    digitalWrite(pin_green, HIGH);   // turn on the green LED
    digitalWrite(pin_red, LOW);    // turn off the red LED
    Serial.print("voltage : "); Serial.println(pinVoltage);
  }

  else if (pinVoltage < 3.1)
  {
    digitalWrite(pin_red, HIGH);   // turn on the red LED
    digitalWrite(pin_green, LOW);    // turn off the green LED
    Serial.print("voltage : "); Serial.println(pinVoltage);
  }
  */
  //-------------------------------------------------------------------------------------------------------------------------------------------------------
  
  //code to set time, and save data
        //set time
        if (M5.BtnA.wasReleased() || M5.BtnA.pressedFor(1000, 200))
        {
          M5.Lcd.clear(BLACK);
          M5.Lcd.setCursor(0, 35);
          M5.Lcd.println("Time setup.");

          delay(1000);

          while(1)
          {
            M5.update();

            if (M5.BtnA.wasReleased() || M5.BtnA.pressedFor(1000, 200))
            {
              if (t >= 20)
              {
                t = 20; //the max total time is 20 seconds
              }

              else
              {
                t++;
              }
      
            M5.Lcd.clear(BLACK);
            M5.Lcd.setCursor(0, 35);
            M5.Lcd.println("Time setup");
            M5.Lcd.println(" ");
            M5.Lcd.print(t); M5.Lcd.println(" seconds.");
      
            }

            else if (M5.BtnB.wasReleased() || M5.BtnB.pressedFor(1000, 200))
            {
              M5.Lcd.clear(BLACK);
              M5.Lcd.setCursor(0, 35);
              M5.Lcd.print("Time set to "); M5.Lcd.print(t); M5.Lcd.println(" seconds.");
              delay(2000);
              break;
            }

            else if (M5.BtnC.wasReleased() || M5.BtnC.pressedFor(1000, 200)) 
            {
              if (t <= 2)
              {
                t = 2; //the min total time is 2 seconds
              }

              else
              {
                t--;
              }
      
              M5.Lcd.clear(BLACK);
              M5.Lcd.setCursor(0, 35);
              M5.Lcd.println("Time setup");
              M5.Lcd.println(" ");
              M5.Lcd.print(t); M5.Lcd.println(" seconds.");
      
            }
          
          }


          //save data
          data_size = t/2; //divide by the period of your choice, data_size = number of values
    
          M5.Lcd.clear(BLACK);
          M5.Lcd.setCursor(0, 35);
          M5.Lcd.println("Saving data...");

          temperature = RequestingData(sensorAddress); 
          dtostrf(temperature, 4, 2, temperature_arr);
          writeFile(SPIFFS, "/sensorData.txt", temperature_arr); //write the file using the first temperature value
          
          for (int i = 1; i < data_size; i++)
          { 
          //appending values until data size is reached
          temperature = RequestingData(sensorAddress);
          dtostrf(temperature, 4, 2, temperature_arr);

          appendFile(SPIFFS, "/sensorData.txt", "\n");
          appendFile(SPIFFS, "/sensorData.txt", temperature_arr);

          readFile(SPIFFS, "/sensorData.txt");

          delay(2000); //we can change the period to suite the required situation
          
          }
    
          M5.Lcd.println("");
          M5.Lcd.println("Done saving data. Connect to send data");
          delay(2000);

          data_present = true; 
        
       }
  

  //-------------------------------------------------------------------------------------------------------------------------------------------------------
  // formatting and sending data as a string

  String str = sensorName + " : " + String(txValue);

    if (deviceConnected) 
    {
      //choosing between RT and saved data
      if(data_present == true)
      {

        while(1)
        {
          M5.update();

          M5.Lcd.clear(BLACK);
          M5.Lcd.setCursor(0, 35);
          M5.Lcd.println("The device contains saved data. Press A to send saved data. Press B to send RT data");
          M5.Lcd.println("");

          if (M5.BtnA.wasReleased() || M5.BtnA.pressedFor(1000, 200))
          {
            send_saved = true;
            send_RT = false;
            break;
          }

          else if (M5.BtnB.wasReleased() || M5.BtnB.pressedFor(1000, 200))
          {
            send_RT = true;
            send_saved = false;
            data_present == false;
            break;
          }
        }
      }

      else
      {
        send_RT = true;
      }

        //code to send RT data
       if(send_RT == true)
       {
        uint8_t arr[30];
        for(int i = 0; i < str.length(); i++)
        {
         arr[i] = str[i];
        }
        Serial.println(str);
        
        //pTxCharacteristic->setValue(&txValue, 1);
        pTxCharacteristic->setValue(arr, str.length());
        pTxCharacteristic->notify();
        //txValue++;

        //requesting and displaying, and sending RT sensor data
        M5.Lcd.clear(BLACK);
        M5.Lcd.setCursor(0, 35);
        M5.Lcd.println("Sending RT data...");
        M5.Lcd.println("");
        M5.Lcd.println(str);
        
        txValue = RequestingData(sensorAddress);
        
        delay(2000); // bluetooth stack will go into congestion, if too many packets are sent
       }


        //code to send saved data
        else if(send_saved == true)
        {
          File file2 = SPIFFS.open("/sensorData.txt");
          
          for (int i = 0; i < data_size; i++)
          {
            
            txValue = file2.readStringUntil('\n'); //txValue reads data line by line from the txt file sensorData.txt
            
            String str_svd = sensorName + " : " + String(txValue);
            
            M5.Lcd.clear(BLACK);
            M5.Lcd.setCursor(0, 35);
            M5.Lcd.println("Sending saved data...");
            M5.Lcd.println(" ");
            M5.Lcd.println(str_svd);

            uint8_t arr_svd[30];
            for(int i = 0; i < str_svd.length(); i++)
            {
              arr_svd[i] = str_svd[i];
            }
            Serial.println(str_svd);
        
            //pTxCharacteristic->setValue(&txValue, 1);
            pTxCharacteristic->setValue(arr_svd, str_svd.length());
            pTxCharacteristic->notify();
            
            delay(2000);
            
          }

          file2.close();

          M5.Lcd.println(" ");
          M5.Lcd.println("Done sending saved data.");
          delay(2000);

          //------------------------
          //deleting the file after sending
          //deleteFile(SPIFFS, "/sensorData.txt");
          //------------------------

          while(1)
          {
            M5.update();
      
            M5.Lcd.clear(BLACK);
            M5.Lcd.setCursor(0, 35);
            M5.Lcd.println("Press Button B to read RT data");

            if (M5.BtnB.wasReleased() || M5.BtnB.pressedFor(1000, 200))
            {
              send_saved = false;
              data_present = false;
              break;
            }
    
          }

        }

        
  }
  //----------------------------------------------------------------------------------------------------------------------------------------------------------

    // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
        delay(500); // give the bluetooth stack the chance to get things ready
        pServer->startAdvertising(); // restart advertising
        Serial.println("start advertising");
        
        M5.Lcd.clear(BLACK);
        M5.Lcd.setCursor(0, 35);
        M5.Lcd.println("start advertising");
        oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
    // do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
}



//-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//Custom functions

float RequestingData(byte sAdd)
{
  Wire.requestFrom(sAdd, 2);
  //The wire can only transfer 1 byte at a time.
  //We need to receive temperature values in 2 bytes since temperature is a float, but the values are received as integers.
  //If the temperature is 25.42 °C, we receive a byte with 25, and another byte with 42.
  //Once we have received all the data we add 25 and 42/100 to get back our float number.

  int data[2];
  float temperature;

  if (Wire.available()) 
  {
    data[0] = Wire.read();
    data[1] = Wire.read();
  }
  
   temperature = data[0] + (float)data[1]/100;

   return temperature;
}


void readFile(fs::FS &fs, const char * path)
{
    Serial.printf("Reading file: %s\r\n", path);

    File file = fs.open(path);
    if(!file || file.isDirectory()){
        Serial.println("- failed to open file for reading");
        return;
    }

    Serial.println("- read from file:");
    while(file.available()){
        Serial.write(file.read());
    }
    file.close();
}

void writeFile(fs::FS &fs, const char * path, const char * message)
{
    Serial.printf("Writing file: %s\r\n", path);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("- failed to open file for writing");
        return;
    }
    if(file.print(message)){
        Serial.println("- file written");
    } else {
        Serial.println("- write failed");
    }
    file.close();
}

void appendFile(fs::FS &fs, const char * path, const char * message)
{
    Serial.printf("Appending to file: %s\r\n", path);

    File file = fs.open(path, FILE_APPEND);
    if(!file){
        Serial.println("- failed to open file for appending");
        return;
    }
    if(file.print(message)){
        Serial.println("- message appended");
    } else {
        Serial.println("- append failed");
    }
    file.close();
}

void renameFile(fs::FS &fs, const char * path1, const char * path2)
{
    Serial.printf("Renaming file %s to %s\r\n", path1, path2);
    if (fs.rename(path1, path2)) {
        Serial.println("- file renamed");
    } else {
        Serial.println("- rename failed");
    }
}

void deleteFile(fs::FS &fs, const char * path)
{
    Serial.printf("Deleting file: %s\r\n", path);
    if(fs.remove(path)){
        Serial.println("- file deleted");
    } else {
        Serial.println("- delete failed");
    }
}
